import unittest
from barycorrpy import get_BC_vel , exposure_meter_BC_vel
import numpy as np
from astropy.time import Time
from barycorrpy import utc_tdb



class Barycorrpy_tests(unittest.TestCase):

    def test_hip_id(self):
        JDUTC = 2458000 # Also accepts float input for JDUTC. Verify scale and format
        result = get_BC_vel(JDUTC=JDUTC,hip_id=8102,lat=-30.169283,longi=-70.806789,alt=2241.9,ephemeris='de430',zmeas=0.0)
        self.assertTrue(np.isclose(a = result[0], b = 15403.9508,atol = 1e-2, rtol = 0), 4)

    def test_hip_id_astropy_obs(self):
        JDUTC = Time(2458000,format='jd',scale='utc')
        result  = get_BC_vel(JDUTC=JDUTC,hip_id=8102,obsname='CTIO',ephemeris='de430')
        self.assertTrue(np.isclose(a = result[0],b = 15403.9608,atol = 1e-2, rtol = 0))

    def test_inputs(self):
        ra=26.0213645867
        dec=-15.9395557246
        obsname=''
        lat=-30.169283
        longi=-70.806789
        alt=2241.9
        epoch = 2451545.0
        pmra = -1721.05
        pmdec = 854.16
        px = 273.96
        rv = 0.0
        zmeas=0.0
        JDUTC=[2458000,2458000.00001,2458000.00002] # Can also enter JDUTC as float instead of Astropy Time Object
        ephemeris='https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp'

        result=get_BC_vel(JDUTC=JDUTC,ra=ra,dec=dec,obsname=obsname,lat=lat,longi=longi,alt=alt,pmra=pmra,
                pmdec=pmdec,px=px,rv=rv,zmeas=zmeas,epoch=epoch,ephemeris=ephemeris,leap_update=True)

        self.assertTrue(np.allclose([result[0][0],result[0][1],result[0][2]],[15407.4860,15407.4723,15407.4586],atol = 1e-2, rtol = 0))

    def test_flux_weighting(self):
        ra=26.0213645867
        dec=-15.9395557246
        obsname=''
        lat=-30.169283
        longi=-70.806789
        alt=2241.9
        epoch = 2451545.0
        pmra = -1721.05
        pmdec = 854.16
        px = 273.96
        rv = 0.0
        zmeas=0.0
        JDUTC=[2458000,2458000.00001,2458000.00002]

        flux = [4.5,1.5,2] # Same number of elements as JDUTC

        result,JDUTCMID,warning4,status4=exposure_meter_BC_vel(JDUTC=JDUTC,expmeterflux = flux,
                ra=ra,dec=dec,obsname=obsname,lat=lat,longi=longi,alt=alt,pmra=pmra,
                pmdec=pmdec,px=px,rv=rv,zmeas=zmeas,epoch=epoch,leap_update=True)

        self.assertTrue(np.isclose(a = result,b = 15407.4765,atol = 1e-2, rtol = 0))


    def test_JDUTC_BJDTDB(self):
        JDUTC=[2458000,2458000.00001,2458000.00002]
        corr_time = utc_tdb.JDUTC_to_BJDTDB(JDUTC,hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9)

        self.assertTrue(np.allclose([corr_time[0]],[2458000.00505211,  2458000.00506211,  2458000.00507211],atol = 1e-7, rtol = 0))

    def test_stellar_predictive(self):
        JDUTC = 2458000
        result5 = get_BC_vel(JDUTC=JDUTC, hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9, ephemeris='de430', zmeas=0.0, predictive=True)

        self.assertTrue(np.isclose(a = result5[0], b = -15403.15938, atol = 1e-2, rtol = 0))


    def test_SolarBC(self):
        JDUTC = 2458000

        result6 = get_BC_vel(JDUTC=2458000, lat=-30.169138888, longi=-70.805888, alt=2379.5, zmeas=0.0, SolSystemTarget='Sun')
        self.assertTrue(np.isclose(a = result6[0], b = 819.4474, atol = 1e-2, rtol = 0))






def main():
    unittest.main()

if __name__ == "__main__":
    main()



from __future__ import division
from __future__ import print_function
import urllib
import os
import datetime
import numpy as np
import sys
import math
import astropy
from astropy.coordinates import EarthLocation
from astropy.coordinates import get_body_barycentric_posvel
from astropy.time import Time

if sys.version_info.major==3:
    from functools import lru_cache
else:
    from functools32 import lru_cache

from . import PINT_erfautils as PINT
from .utils import get_stellar_data,find_hip
from .PhysicalConstants import *

fpath=os.path.join(os.path.dirname(__file__),'data')

def staleness_check(file_time,now):
    '''
    Check whether the leap second file needs to be updated.

    Leap second updates generally announced shortly after December 31st or June 30th. Thus assuming a month lag for the list to be updated,
    check if a February 1st or August 1st has passed between when the file was downloaded and when the check is being run.

    INPUT:
        Both inputs are in datetime format
        file_time : Time the leap second file was last downloaded. Stored in leap_log.txt.
        now : UTC time right now.


    Return 1 if need to update, else return 0.

    '''
    if file_time.month > 8 :
        year=file_time.year+1
    else: year=file_time.year

    if (file_time < datetime.datetime(year,2,1) <= now) or (file_time < datetime.datetime(year,9,1) <= now):
        return 1
    else: return 0


def leap_download(ls_fpath, log_fpath):
    """
    Download the leap second file and update the log file.
    INPUT:
        ls_fpath : Path to where the file will be saved.
        log_fpath : Path to where the log file is created.

    """

    url = 'http://maia.usno.navy.mil/ser7/tai-utc.dat'

    if sys.version_info.major == 3:
        try:
            request = urllib.request.urlopen(url, timeout=10)  # timeout in 10 seconds if no response from server
            with open(ls_fpath, 'w') as f:
                f.write(request.read())
            flag = 0
            with open(log_fpath, 'w') as f:
                f.write(str(datetime.datetime.utcnow()))  # Write date of download in log file
        except (urllib.error.URLError, IOError):
            flag = 1
    else:
        import urllib2
        try:
            # should this be urllib2.urlretrieve?
            urllib2.urlretrieve(url, ls_fpath)
            flag = 0
            with open(log_fpath, 'w') as f:
                f.write(str(datetime.datetime.utcnow()))  # Write date of download in log file
        except (urllib2.HTTPError, IOError):
            flag = 1

    return flag


def leap_manage(utctime,fpath,leap_update):
    '''

    Calculates the offset between UTC and TAI from the leap second file.
    Also, checks for 'staleness' (how old is the file) of the leap second file.
    When the leap second file gets about 6 months old, it will update the leap second file, to check if there has been
    a new leap second announced.


    INPUT:
        Enter UTC time as Astropy Time Object. In UTC Scale.

        fpath : Path to where the file would be saved.
        leap_update : If True, when the leap second file is more than 6 months old it will attempt to download a new one.
                    If False, then will just give a warning message.

    OUTPUT:
        tai_utc : Leap second offset between TAI and UTC (Scalar)
        warning, error : Warning and Error messages

        Offset in seconds, between TAI and UTC.
    '''

    warning=[]
    error=[]

    # Location of saved leap second file
    ls_fpath=os.path.join(fpath,'leap_sec.txt')

    # Location of log file to record how old is the leap second file
    log_fpath=os.path.join(fpath,'leapsec_log.txt')

    # If log or leap file does not exist, then download the leap second file and create a log file
    if (os.path.isfile(log_fpath)==False or os.path.isfile(ls_fpath)==False):
        if leap_update==True:
            flag=leap_download(ls_fpath=ls_fpath,log_fpath=log_fpath)
            if flag==0:
                warning+=['WARNING JD = '+str(utctime.jd)+' : Downloaded leap second file from http://maia.usno.navy.mil/ser7/tai-utc.dat ']
            else:
                error+=['ERROR : JD = '+str(utctime.jd)+' :  Unable to download leap second file. Check internet connection or leap_dir parameter']
                return 0,warning,error
        else:
            error+=['ERROR : JD = '+str(utctime.jd)+' :  LEAP SECOND FILE / LOG FILE DOES NOT EXIST. Please set leap_update = True to update file. Corrections may not be accurate ']
            return 0,warning,error


    # If file exists then check if need to update
    else:

        file_time = _read_log(log_fpath)
        now=datetime.datetime.utcnow()  # Current time

        if file_time < utctime.datetime:  # Check for file staleness if test time is after the last file update time

            # Check if need to update file
            if staleness_check(file_time,now)==1:
                warning+=['WARNING : JD = '+str(utctime.jd)+' :  Need to update leap second file. Corrections may not be accurate to 1 cm/s with old file']
                if leap_update==True:
                    flag=leap_download(ls_fpath=ls_fpath,log_fpath=log_fpath)
                    if flag==0:
                        warning+=['JD = '+str(utctime.jd)+' : Downloaded leap second file from http://maia.usno.navy.mil/ser7/tai-utc.dat ']
                    else:
                        error+=['ERROR : JD = '+str(utctime.jd)+' :  Unable to download leap second file. Check internet connection or leap_dir parameter']
                        return 0,warning,error

                else:
                    error+=['ERROR : JD = '+str(utctime.jd)+' :  Leap second file should be updated. Set leap_update = True to download file']
                    return 0,warning,error


    jd, offset = _read_leapfile(ls_fpath)

    if type(utctime)!=astropy.time.core.Time: # Check if it is in Astropy Time Object
        print("Input Time should be as Astropy Time Object")
        raise TypeError("Input Time should be as Astropy Time Object")
    else:
        JDUTC=utctime.jd

    tai_utc=offset[np.max(np.where(JDUTC>=np.array(jd)))] # Leap second offset value to convert UTC to TAI

    return tai_utc  , warning , error


def JDUTC_to_JDTDB(utctime,leap_update=True,fpath=fpath):
    '''
    Convert JDUTC to JDTDB (Barycentric Dynamical Time)
    INPUT:
        utctime : Enter UTC time as Astropy Time Object. In UTC Scale.
        fpath : Path to where the file would be saved. Default is script directory.
        leap_update : If True, when the leap second file is more than 6 months old it will attempt to download a new one.
                      If False, then will just give a warning message.  Default is True.

    OUTPUT:
        JDTDB : Julian Date Barycentric Dynamic Time (Astropy Time object)
        JDTT: Julian Date Terrestrial Dynamic time (Astropy Time object)
        warning,error : Warning and Error message if any from the routine




    '''

    JDUTC=utctime.jd
    if JDUTC<2441317.5 :
        return utctime.tdb,utctime.tt,['WARNING : JD = '+str(utctime.jd)+' :  Precise leap second history is not maintained here for before 1972. Defaulting to Astropy for time conversion. Corrections maybe inaccurate.'],[]

    # Call function to check leap second file and find offset between UTC and TAI.
    tai_utc,warning,error=leap_manage(utctime=utctime,fpath=fpath,leap_update=leap_update)

    if utctime.scale != 'utc':
        error+['ERROR : JD = '+str(utctime.jd)+' : Please input time in UTC scale']
        if utctime.scale == 'tdb':
            return utctime.tdb,utctime.tt,['WARNING : JD = '+str(utctime.jd)+' :  UTC Time scale not used. Defaulting to Astropy for time conversion. Corrections maybe inaccurate.'],[]

    if tai_utc==0:
        return utctime.tdb,utctime.tt,warning,error+['ERROR : JD = '+str(utctime.jd)+' :  Unable to maintain leap second file. Defaulting to AstroPy version of leap seconds.']

    check_time=utctime.datetime


    # Add leap seconds to convert UTC to TAI
    new_tai=check_time+datetime.timedelta(seconds=tai_utc)  # Add offset and convert to TAI
    new_tt=new_tai+datetime.timedelta(seconds=32.184)  # Add 32.184 to convert TAI to TT

    g=(357.53+0.9856003*( JDUTC - 2451545.0 ))*np.pi/180.  # Earth's mean anomaly

    TDB = new_tt+datetime.timedelta(seconds=0.001658*np.sin(g)+0.000014*np.sin(2*g)) # TT to TDB

    JDTT=Time(new_tt,scale='tt',format='datetime')
    JDTT.format='jd'
    JDTDB = Time(TDB,scale='tdb',format='datetime')
    JDTDB.format='jd'

    return JDTDB,JDTT,warning,error






def JDUTC_to_BJDTDB(JDUTC,
       starname = '', hip_id=None, ra=None, dec=None, epoch=None, pmra=None, pmdec=None, px=None, rv=None,
       obsname='', lat=0., longi=0., alt=0.,
       ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True):

    '''
    Time conversion between JDUTC to BJDTDB. See Eastman et al. (2010)
    This code is precise to about 200 us.

    Calling procedure for JDUTC_to_BJDTDB. Accepts vector time object (i.e., multiple observation JD values).

    INPUT:
        JDUTC : Can enter multiple times in Astropy Time object or as float. Will loop through and find barycentric velocity correction corresponding to those times.
                In UTC Scale. If using float, be careful about format and scale used.
        starname : Name of target. Will query SIMBAD database.
                                OR / AND
        hip_id : Hipparcos Catalog ID. (Integer) . Epoch will be taken to be Catalogue Epoch or J1991.25
                If specified then ra,dec,pmra,pmdec,px, and epoch need not be specified.
                                OR / AND
        ra, dec : RA and Dec of star [degrees].
        epoch : Epoch of coordinates in Julian Date. Default is J2000 or 2451545.
        pmra : Proper motion in RA [mas/year]. Eg. PMRA = d(RA)/dt * cos(dec). Default is 0.
        pmdec : Proper motion in Dec [mas/year]. Default is 0.
        px : Parallax of target [mas]. Default is 0.

        obsname : Name of Observatory as defined in Astropy EarthLocation routine. Can check list by EarthLocation.get_site_names().
                  If obsname is not used, then can enter lat,long,alt.
                                OR
        lat : Latitude of observatory in [degrees]. North (+ve) and South (-ve).
        longi : Longitude of observatory [degrees]. East (+ve) and West (-ve).
        alt : Altitude of observatory [m].

        rv : Radial Velocity of Target [m/s]. Default is 0.
        ephemeris : Name of Ephemeris to be used. List of Ephemeris as queried by jplephem. Default is DE430.
                    For first use Astropy will download the Ephemeris ( for DE430 ~100MB). Options for ephemeris inputs are
                    ['de432s','de430',
                    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de423_for_mercury_and_venus/de423.bsp',
                    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp']
        leap_dir : Directory where leap seconds file will be saved and maintained (STRING). Eg. '/Users/abc/home/savehere/'. Default is script directory.
        leap_update : If True, when the leap second file is more than 6 months old will attempt to download a new one.
                      If False, then will just give a warning message. Default is True.

    OUTPUT:
        corr_time : BJDTDB time
        warning : Warning and Error message from the routine.
        status : Status regarding warning and error message. Returns the following -
                0 - No warning or error.
                1 - Warning message.
                2 - Error message.

    Example:
    >>> from astropy.time import Time
    >>> JDUTC = Time(2458000, format='jd', scale='utc')
    >>> utc_tdb.JDUTC_to_BJDTDB(JDUTC,hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9)
    (array([ 2458000.00505211]), [[], []], 0)

    '''

    corr_time = []
    warning = []
    error = []
    status = 0

    # Check for JDUTC type
    if type(JDUTC)!=Time:
         warning += [['Warning: Float JDUTC entered. Verify time scale (UTC) and format (JD)']]
         JDUTC=Time(JDUTC, format='jd', scale='utc')

    if JDUTC.isscalar:
        JDUTC = Time([JDUTC])

    star_par = {'ra':ra,'dec':dec,'pmra':pmra,'pmdec':pmdec,'px':px,'rv':rv,'epoch':epoch}
    star_simbad = {'ra':None,'dec':None,'pmra':None,'pmdec':None,'px':None,'rv':None,'epoch':None}
    star_hip = {}
    star_zero = {'ra':0.,'dec':0.,'pmra':0.,'pmdec':0.,'px':0.,'rv':0.,'epoch':2451545.0}
    star_output = {}


    if starname:
        star_simbad,warning1 = get_stellar_data(starname)
        warning += warning1
    if hip_id:
        if starname:
            warning += ['Warning: Querying SIMBAD and Hipparcos Catalogue']
        star_hip = find_hip(hip_id)


    star_output = star_simbad.copy()
    star_output.update({k:star_hip[k] for k in star_hip if star_hip[k] is not None})
    star_output.update({k:star_par[k] for k in star_par if star_par[k] is not None})
    star_output.update({k:star_zero[k] for k in star_zero if star_output[k] is None})
    warning+=['Following are the stellar positional parameters being used - ',star_output]


    if obsname:
        loc = EarthLocation.of_site(obsname)
        lat = loc.lat.value
        longi = loc.lon.value
        alt = loc.height.value
        warning += [['Warning: Taking observatory coordinates from Astropy Observatory database. Verify precision. Latitude = %f  Longitude = %f  Altitude = %f'%(lat,longi,alt)]]
    else:
        loc = EarthLocation.from_geodetic(longi, lat, height=alt)

    for jdutc in JDUTC:
        a = _JDUTC_to_BJDTDB(JDUTC=jdutc,
                 loc=loc,
                 ephemeris=ephemeris, leap_dir=leap_dir, leap_update=leap_update,**star_output)
        corr_time.append(a[0])
        warning.append(a[1])
        error.append(a[2])


    # Status messages to check for warning or error
    if not all(corr_time): error += ['Check inputs. Error in code']
    if any(error):   status |= 2
    if any(warning): status |= 1
    # Convert corrected from list to numpy array
    corr_time = np.array(corr_time)

    return corr_time, warning+error, status



def _JDUTC_to_BJDTDB(JDUTC,
    ra=0.0, dec=0.0, epoch=2451545.0, pmra=0.0, pmdec=0.0, px=0.0, rv=0.0,
    loc=None,
    ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True):

    '''
    Time conversion between JDUTC to BJDTDB. See Eastman et al. (2010)
    This code is precise to about 200 us.

    See JDUTC_to_BJDTDB() for parameter description.

    '''

    # Convert times to obtain TDB and TT
    JDTDB, JDTT, warning, error = JDUTC_to_JDTDB(JDUTC, fpath=leap_dir, leap_update=leap_update)
    clock_corr = (JDTDB.jd - JDUTC.jd) * 86400.

    ##### NUTATION, PRECESSION, ETC. #####

    r_pint, v_pint = PINT.gcrs_posvel_from_itrf(loc, JDUTC, JDTT)

    r_eci = r_pint[0]  # [m]
    v_eci = v_pint[0]  # [m/s]

    ##### EPHEMERIDES #####

    earth_geo = get_body_barycentric_posvel('earth', JDTDB, ephemeris=ephemeris) # [km]
    r_obs = r_eci + earth_geo[0].xyz.value*1000. # [m]
    v_geo = earth_geo[1].xyz.value*1000./86400.  # [m/s]

    # Relativistic Addition of Velocities
    v_obs = (v_eci+v_geo) / (1.+v_eci*v_geo/c**2) # [m/s]

    # calculate the Einstein delay relative to the geocenter
    # (TDB accounts for Einstein delay to geocenter)
    einstein_corr = np.sum(r_eci*v_geo)/(c*c)


    ##### Convert Star RA DEC to R0hat vector #####

    r0hat = np.array([math.cos(ra*np.pi/180.)*math.cos(dec*np.pi/180.),
                      math.sin(ra*np.pi/180.)*math.cos(dec*np.pi/180.),
                                              math.sin(dec*np.pi/180.)])
    # Eq 14 to 17
    up = [0., 0., 1.]
    east = np.cross(up, r0hat)
    east = east / math.sqrt(sum(east*east))
    north = np.cross(r0hat, east)
    mu = (pmra*east+pmdec*north)/pctoau/1000 # Divided by 1000 since the Proper motion is in milli-arcseconds.


    ##### Convert Star RA DEC to R0hat vector #####

    r0hat = np.array([math.cos(ra*np.pi/180.)*math.cos(dec*np.pi/180.),
                      math.sin(ra*np.pi/180.)*math.cos(dec*np.pi/180.),
                                              math.sin(dec*np.pi/180.)])
    # Eq 14 to 17
    up = [0., 0., 1.]
    east = np.cross(up, r0hat)
    east = east / math.sqrt(sum(east*east))
    north = np.cross(r0hat, east)
    mu = (pmra*east+pmdec*north)/pctoau/1000 # Divided by 1000 since the Proper motion is in milli-arcseconds.


    ##### Stellar position corrected for motion #####

    epoch0 = 2000. + (epoch-2451545.)/365.25
    yearnow = 2000. + (JDTDB.jd-2451545.)/365.25

    T = yearnow - epoch0                           # [years]
    vpi = rv/1.e3 * kmstoauyr * (px/1.e3/pctoau)   # [rad/yr]
    vel = mu + vpi*r0hat                           # [rad/yr] (m in AA)
    r = r0hat + vel*T                              # [rad]    (p1 in AA)
    rhat = r / math.sqrt(sum(r*r))

    # Geometric Correction
    geo_corr = np.sum(r_obs*rhat)/c


    delta_t = geo_corr + clock_corr + einstein_corr
    result = JDUTC.jd+delta_t/86400.

    return result, warning, error

@lru_cache(maxsize=200)
def _read_log(log_fpath):
    with open(log_fpath,'r') as f:
        file_time=datetime.datetime.strptime(f.readline() , '%Y-%m-%d %H:%M:%S.%f') # Read date of download in log file
    return file_time

@lru_cache(maxsize=200)
def _read_leapfile(ls_fpath):
    """
    Parse in leap second file
    """

    f=open(ls_fpath,'r')
    jd=[]
    offset=[]

    for line in f:
        a=(line.split())
        jd.append(float(a[4]))
        offset.append(float(a[6]))
    f.close()

    return jd, offset



from .utc_tdb import *
from .PINT_erfautils import *
from .utils import *
from .tests import *
from .sample_script import *
from .barycorrpy import get_BC_vel, BCPy, exposure_meter_BC_vel
from .SolarSystemBC import SolarBarycentricCorrection, ReflectedLightBarycentricCorrection

__version__ = '0.3.4'



from __future__ import division
from __future__ import print_function
#import de423    #https://pypi.python.org/pypi/jplephem
from astropy.coordinates import EarthLocation
from astropy.coordinates import get_body_barycentric_posvel, get_body_barycentric
from astropy.time import Time
import math
import numpy as np
import os


from . import find_hip
from . import PINT_erfautils as PINT
from . import utc_tdb
from .SolarSystemBC import SolarBarycentricCorrection, ReflectedLightBarycentricCorrection
from .utils import flux_weighting, get_stellar_data, CalculatePositionVector
from .PhysicalConstants import *

### Need to install jplephem ###
#de430 is 100 MB in size


def get_BC_vel(JDUTC,
       starname= '', hip_id=None, ra=None, dec=None, epoch=None, pmra=None, pmdec=None, px=None, rv=None,
       obsname='', lat=0., longi=0., alt=0., zmeas=0.,
       ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True, predictive=False,
       SolSystemTarget=None, HorizonsID_type='smallbody'):
    '''
    Barycentric Velocity Correction at the 1 cm/s level, as explained in Wright & Eastman, 2014.
    Calling procedure for barycorrpy. Accepts vector time object (i.e., multiple observation JD values).

    Note: Can be used for stellar observations, or for observations of the Sun or other Solar System objects.

    INPUT:
        JDUTC : Can enter multiple times in Astropy Time object or as float. Will loop through and find barycentric velocity correction corresponding to those times.
                In UTC Scale. If using float, be careful about format and scale used.
        starname : Name of target. Will query SIMBAD database.
                                OR / AND
        hip_id : Hipparcos Catalog ID. (Integer). Epoch will be taken to be Catalogue Epoch or J1991.25
                If specified then ra,dec,pmra,pmdec,px, and epoch need not be specified.
                                OR / AND
        ra, dec : RA and Dec of star [degrees].
        epoch : Epoch of coordinates in Julian Date. Default is J2000 or 2451545.
        pmra : Proper motion in RA [mas/year]. Eg. PMRA = d(RA)/dt * cos(dec). Default is 0.
        pmdec : Proper motion in Dec [mas/year]. Default is 0.
        px : Parallax of target [mas]. Default is 0.
        rv : Radial Velocity of Target [m/s]. Default is 0. This is the bulk RV (systemic) of the target at the ~100 km/s precision.
             Can be ignored for most targets but for high velocity stars. This does not include the barycentric velocity and is
             only required to correct for proper motion and secular acceleration.

        obsname : Name of Observatory as defined in Astropy EarthLocation routine. Can check list by EarthLocation.get_site_names().
                  If obsname is not used, then can enter lat,long,alt.
                                OR
        lat : Latitude of observatory in [degrees]. North (+ve) and South (-ve).
        longi : Longitude of observatory [degrees]. East (+ve) and West (-ve).
        alt : Altitude of observatory [m].

        zmeas : Measured redshift (e.g., the result of cross correlation with template spectrum). Default is 0.
                The redshift measured by the spectrograph before any barycentric correction. Therefore zmeas includes the barycentric
                velocity of the observatory.
        ephemeris : Name of Ephemeris to be used. List of Ephemeris as queried by jplephem. Default is DE430.
                    For first use Astropy will download the Ephemeris ( for DE430 ~100MB). Options for ephemeris inputs are
                    ['de432s','de430',
                    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de423_for_mercury_and_venus/de423.bsp',
                    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp']
        leap_dir : Directory where leap seconds file will be saved and maintained (STRING). Eg. '/Users/abc/home/savehere/'. Default is
                script directory.
        leap_update : If True, when the leap second file is more than 6 months old will attempt to download a new one.
                      If False, then will just give a warning message. Default is True.
        SolSystemTarget : When running barycentric correction for a stellar target, Target = None. Default value = None
                To correct for Solar RV observations set Target = 'Sun', for reflected light observations, see below.

                For Solar observations:
                Only inputs required are -
                    1. JDUTC
                    2. loc
                    3. zmeas
                    4. ephemeris
                    5. leap_dir
                    6. leap_update
                    7. predictive

                For Reflected light observations:
                    For observations of the Moon: SolSystemTarget='301' and HorizonsID_type='majorbody'
                    For observations of asteroids: SolSystemTarget=ObjectName, where the ObjectName is queriable by Horizons and HorizonsID_type='smallbody' [default]

        HorizonsID_type : Refers to the Horizons id type to identify the object type, and is required for the reflected light observations.
        > 'smallbody' refers to asteroid or comet and is the default. Use for Asteroids.
        > 'majorbody' refers to planets or satellites. Use for Moon.

        predictive : If True, then instead of returning v_true, returns v_predicted.
                Default: False, and return is v_true from Wright and Eastman (2014)


    OUTPUT:
        FOR STELLAR OBSERVATIONS (not the Sun)
        If SolSystemTarget == None
            If predictive = False
                v_true : The barycenter-corrected RV [m/s] as defined in Wright & Eastman, 2014.
                NOTE: This is not just the barycentric velocity that can be subtracted directly from the measured RV.
                    The measured RV must be entered in the code as zmeas. This is because the relativistic cross product between zbary and zmeas is
                    required. This matters at ~ m/s level and hence must be included.

                The formula used is ztrue = ((1.+ zb)*(1.+ zmeas) - 1.)
                Therefore if zmeas is set to 0, then ztrue = zb. The velocities are just the redshift (z) x speed of light (c).
            Else if predictive = True
                v_predicted
                In this predictive mode, the expected RV shift measured by the instrument is output.
                NOTE: zpredicted = zmeas for an ideal noiseless observation, where the RV from the star (ztrue) is zero (or a flat line vs time)


            warning : Warning and Error message from the routine.
            status : Status regarding warning and error message. Returns the following -
                    0 - No warning or error.
                    1 - Warning message.
                    2 - Error message.

        FOR SOLAR OBSERVATIONS
        If SolSystemTarget == 'Sun'
            If predictive = False
                v_true: The true radial velocity of the Sun for an observer at the observatory but in an inertial frame not moving.
                    If zmeas is included to show the measured absolute redshift for the Sun as measured by an instrument,
                    then in this formulation, v_true will show the motion of the Sun,
                    which is mostly dominated by the synodic period of Jupiter as seen from Earth.
                    Our barycentric correction includes the gravitational redshift of the Sun, and therefore the true
                    velocity (or redshift) is centered at 0 m/s.
            Else if predictive = True
                v_predicted: Ideal redshift measured for the Sun from Earth for given location and time.
                    This output returns the theoretical prediction for the redshift which includes the barycentric component.
                    This will be the measurement of a noiseless RV instrument observing the Sun.
        Else:
            Computing the barycentric corrections for reflected light observations of a target in the Solar system
            If predictive = False
                v_true: The true radial velocity of the SolSystemTarget for an observer at the observatory but in an inertial frame not moving.
                    If zmeas is included to show the measured absolute redshift for the SolSystemTarget as measured by an instrument,
                    then in this formulation, v_true will show the motion of the SolSystemTarget.
            Else if predictive = True
                v_predicted: Ideal redshift measured for the SolSystemTarget from Earth for given location and time.
                    This output returns the theoretical prediction for the redshift which includes the barycentric component.
                    This will be the measurement of a noiseless RV instrument observing the SolSystemTarget.

            The formula used is ztrue = ((1.+zb)*(1.+zmeas)-1.)
            Therefore if zmeas is set to 0, then ztrue = zb. The velocities are just the redshift (z) x speed of light (c).

            warning : Warning and Error message from the routine.
            status : Status regarding warning and error message. Returns the following -
                    0 - No warning or error.
                    1 - Warning message.
                    2 - Error message.


    Example:
    >>> from astropy.time import Time
    >>> JDUTC = Time(2458000, format='jd', scale='utc')
    >>> get_BC_vel(JDUTC, hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9)
    (array([15403.95089287]), [[], []], 0)

    '''
    warning = []
    error = []
    status = 0

    # Check for JDUTC type
    if type(JDUTC)!=Time:
         warning += [['Warning: Assuming input time is in UTC JD. Pass astropy Time object to avoid this warning.']]
         JDUTC = Time(JDUTC, format='jd', scale='utc')

    if JDUTC.isscalar:
        JDUTC = Time([JDUTC])
    else:
       if np.size(zmeas)>1 and np.size(zmeas)!=np.size(JDUTC):
          error+= [['Error: Size mismatch. JDUTC is a vector, zmeas must also be a vector of same length corresponding to those dates']]
          raise IndexError('Error: Size mismatch. JDUTC is a vector, zmeas must be a vector of same length corresponding to those dates')

    if obsname:
        loc = EarthLocation.of_site(obsname)
        lat = loc.lat.value
        longi = loc.lon.value
        alt = loc.height.value
        warning += ['Warning: Taking observatory coordinates from Astropy Observatory database. Verify precision. Latitude = %f  Longitude = %f  Altitude = %f'%(lat,longi,alt)]
    else:
        loc = EarthLocation.from_geodetic(longi, lat, height=alt)

    # check if we need to update the leap second file up front using the maximum
    # of the JDUTC array. Then we turn it off regardless of what the user had set
    if leap_update:
        maxutc = max(JDUTC)
        # Call function to check leap second file and find offset between UTC and TAI.
        tai_utc, warning1, error1 = utc_tdb.leap_manage(utctime=maxutc, fpath=leap_dir, leap_update=leap_update)
        warning += warning1
        error += error1

    # STELLAR TARGET #
    if SolSystemTarget is None:
        # Running correction for stellar observation
        star_par = {'ra':ra,'dec':dec,'pmra':pmra,'pmdec':pmdec,'px':px,'rv':rv,'epoch':epoch}
        star_simbad = {}
        star_hip = {}
        star_zero = {'ra':0.,'dec':0.,'pmra':0.,'pmdec':0.,'px':0.,'rv':0.,'epoch':2451545.0}
        star_output = {}

        if starname:
            star_simbad,warning1 = get_stellar_data(starname)
            warning += warning1
        if hip_id:
            if starname:
                warning += ['Warning: Querying SIMBAD and Hipparcos Catalogue']
            star_hip = find_hip(hip_id)

        star_output = star_simbad.copy()
        star_output.update({k:star_hip[k] for k in star_hip if star_hip[k] is not None})
        star_output.update({k:star_par[k] for k in star_par if star_par[k] is not None})
        star_output.update({k:star_zero[k] for k in star_output if star_output[k] is None})
        warning+=['Following are the stellar positional parameters being used - ',star_output]
        vel = []

        for jdutc,zm in zip(JDUTC,np.repeat(zmeas,np.size(JDUTC)/np.size(zmeas))):
            a = BCPy(JDUTC=jdutc,
                     zmeas=zm,
                     loc=loc,
                     ephemeris=ephemeris, leap_dir=leap_dir, leap_update=False,
                     predictive=predictive, **star_output)

            vel.append(a[0])
            warning.append(a[1])
            error.append(a[2])

    ## SOLAR OBSERVATIONS ##
    elif SolSystemTarget == 'Sun' or SolSystemTarget == 'Sol':
        vel = []
        for jdutc,zm in zip(JDUTC,np.repeat(zmeas,np.size(JDUTC)/np.size(zmeas))):
            a = SolarBarycentricCorrection(JDUTC=jdutc, loc=loc, zmeas=zm,
                    ephemeris=ephemeris, leap_dir=leap_dir, leap_update=False, predictive=predictive)
            vel.append(a[0])
            warning.append(a[1])
            error.append(a[2])

    else:
        print("Reflected light observations of solar system objects include complexities not accounted for in this calculation, such as rotation, phase effects, and finite source effects.  See Wright & Kanodia (2020)")
        warning+=["Reflected light observations of solar system objects include complexities not accounted for in this calculation, such as rotation, phase effects, and finite source effects.  See Wright & Kanodia (2020)"]
        vel = []
        for jdutc,zm in zip(JDUTC,np.repeat(zmeas,np.size(JDUTC)/np.size(zmeas))):
            a = ReflectedLightBarycentricCorrection(SolSystemTarget=SolSystemTarget, JDUTC=jdutc, loc=loc, zmeas=zm, HorizonsID_type=HorizonsID_type,
                    ephemeris=ephemeris, leap_dir=leap_dir, leap_update=False, predictive=predictive)
            vel.append(a[0])
            warning.append(a[1])
            error.append(a[2])
    # Status messages to check for warning or error
    if not all(vel): error += ['Check inputs. Error in code']
    if any(error):   status |= 2
    if any(warning): status |= 1
    # Convert velocity from list to numpy array
    vel = np.array(vel)

    return vel, warning+error, status






def BCPy(JDUTC,
    ra=0.0, dec=0.0, epoch=2451545.0, pmra=0.0, pmdec=0.0, px=0.0, rv=0.0, zmeas=0.0,
    loc=None,
    ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True, predictive=False):
    '''
    Barycentric Velocity Correction at the 1 cm/s level, as explained in Wright & Eastman, 2014.

    See get_BC_vel() for parameter description.

    '''

    # Convert times to obtain TDB and TT
    JDTDB, JDTT, warning, error = utc_tdb.JDUTC_to_JDTDB(JDUTC, fpath=leap_dir, leap_update=leap_update)

    ##### NUTATION, PRECESSION, ETC. #####

    r_pint, v_pint = PINT.gcrs_posvel_from_itrf(loc, JDUTC, JDTT)

    r_eci = r_pint[0]  # [m]
    v_eci = v_pint[0]  # [m/s]

    ##### EPHEMERIDES #####

    earth_geo = get_body_barycentric_posvel('earth', JDTDB, ephemeris=ephemeris) # [km]
    PosVector_EarthSSB = r_eci + earth_geo[0].xyz.value*1000. # [m]
    v_geo = earth_geo[1].xyz.value*1000./86400.  # [m/s]

    # Relativistic Addition of Velocities
    VelVector_EarthSSB = (v_eci+v_geo) / (1.+np.sum(v_eci*v_geo)/c**2) # [m/s]
    BetaEarth = VelVector_EarthSSB / c

    ##### Convert Star RA DEC to R0hat vector #####

    r0hat = np.array([math.cos(ra*np.pi/180.)*math.cos(dec*np.pi/180.),
                      math.sin(ra*np.pi/180.)*math.cos(dec*np.pi/180.),
                                              math.sin(dec*np.pi/180.)])
    # Eq 14 to 17
    up = [0., 0., 1.]
    east = np.cross(up, r0hat)
    east = east / math.sqrt(np.sum(east*east))
    north = np.cross(r0hat, east)
    mu = (pmra*east+pmdec*north)/pctoau/1000 # Divided by 1000 since the Proper motion is in milli-arcseconds.


    ##### Stellar position corrected for motion #####

    epoch0 = 2000. + (epoch-2451545.)/365.25
    yearnow = 2000. + (JDTDB.jd-2451545.)/365.25

    T = yearnow - epoch0                           # [yr]
    vpi = rv/1.e3 * kmstoauyr * (px/1.e3/pctoau)   # [rad/yr]
    vel = mu + vpi*r0hat                           # [rad/yr] (m in AA)
    r = r0hat + vel*T                              # [rad]    (p1 in AA)
    rhat = r / math.sqrt(np.sum(r*r))

    # Parallax correction #
    if px>0:
        rho = 1000.*rhat/px*pctoau - PosVector_EarthSSB/AU # [AU]
        rhohat = rho / math.sqrt(np.sum(rho*rho)) # Unitless
        r0 = 1000./px*pctoau*AU # [m]
        BetaStar = r0*mu/c/year + rv*r0hat/c

        zlighttravel = rv*r0*np.sum(mu*mu)*T/(year*c*c)

    else:
        rhohat = rhat
        BetaStar = [0., 0., 0.]
        zlighttravel = 0.


    ##### Calculate Gravitaional Redshift due to Solar system bodies #####

    Sum_GR = 0.
    zshapiro = 0.

    for ss_body in ss_bodies:
        if ss_body == 'Earth':
            PosVector_SSObject = earth_geo[0].xyz.value*1000. # [m]
        else:
            jplephem = get_body_barycentric(ss_body, JDTDB, ephemeris=ephemeris)
            PosVector_SSObject = jplephem.xyz.value*1000. # [m]

        # Vector from object barycenter to Observatory
        PosVector_EarthSSObject, PosMag_EarthSSObject, PosHat_EarthSSObject = CalculatePositionVector(r1=PosVector_EarthSSB, r2=PosVector_SSObject)

        # Add Shapiro Delay
        a = np.dot((rhohat-np.dot(PosHat_EarthSSObject,rhohat)*PosHat_EarthSSObject), BetaEarth)
        zshapiro += -2.*GM[ss_body]*a / ((c*c)*PosMag_EarthSSObject*(1+np.dot(PosHat_EarthSSObject, rhohat)))   # Eq 27

        if PosMag_EarthSSObject:
            Sum_GR += GM[ss_body] / PosMag_EarthSSObject  # [(m/s)^2]

    zgravity = 1./(1+Sum_GR/(c*c)) - 1


    ##### Determine the Barycentric RV correction (Eq 28) #####

    GammaEarth = 1. / math.sqrt(1.- np.sum(BetaEarth**2))

    zb = -1. - zshapiro - zlighttravel + \
        GammaEarth*(1+np.dot(BetaEarth, rhohat))*(1+np.dot(r0hat, BetaStar))/((1.+np.dot(BetaStar, rhohat))*(1.+zgravity)) # Eq 28

    if not predictive:
        v_final = c * ((1.+zb)*(1.+zmeas)-1.)  # [m/s]
    else:
        v_final = ((1/(1.+zb)) - 1) * c # [m/s]


    return v_final, warning, error


def exposure_meter_BC_vel(JDUTC, expmeterflux,
       starname = '', hip_id=None, ra=None, dec=None, epoch=None, pmra=None, pmdec=None, px=None, rv=None,
       obsname='', lat=0., longi=0., alt=0., zmeas=0.,
       ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True,
       SolSystemTarget=None, HorizonsID_type='smallbody', predictive=False):

    '''
    Calculate Barycentric velocity weighted by flux from exposure meter to account for long exposure time.
    Enter JDUTC and expmeterflux from exposure meter readings to calculate barycentric velocity correction for exposure.

    INPUT:
        JDUTC : Can enter multiple times in Astropy Time object. In UTC Scale.
                Can also accept list of float JDUTC times. Be cautious about scale used.
        expmeterflux : Array or List of exposure meter fluxes corresponding to each JDUTC.
                       The resultant barycentric correction will be calculated at each JDUTC and then weighted by the exposure meter fluxes. (See Landoni 2014)
        starname : Name of target. Will query SIMBAD database.
                                OR / AND
        hip_id : Hipparcos Catalog ID. (Integer) . Epoch will be taken to be Catalogue Epoch or J1991.25
                If specified then ra,dec,pmra,pmdec,px, and epoch need not be specified.
                                OR / AND
        ra, dec : RA and Dec of star [degrees].
        epoch : Epoch of coordinates in Julian Date. Default is J2000 or 2451545.
        pmra : Proper motion in RA [mas/year]. Eg. PMRA = d(RA)/dt * cos(dec). Default is 0.
        pmdec : Proper motion in Dec [mas/year]. Default is 0.
        px : Parallax of target [mas]. Default is 0.
        rv : Radial Velocity of Target [m/s]. Default is 0. This is the bulk RV (systemic) of the target at the ~100 km/s precision.
             Can be ignored for most targets but for high velocity stars. This does not include the barycentric velocity and is
             only required to correct for proper motion and secular acceleration.

        obsname : Name of Observatory as defined in Astropy EarthLocation routine. Can check list by EarthLocation.get_site_names().
                  If obsname is not used, then can enter lat,long,alt.
                                OR
        lat : Latitude of observatory in [degrees]. North (+ve) and South (-ve).
        longi : Longitude of observatory [degrees]. East (+ve) and West (-ve).
        alt : Altitude of observatory [m].


        zmeas : Measured redshift (e.g., the result of cross correlation with template spectrum). Default is 0.
                The redshift measured by the spectrograph before any barycentric correction. Therefore zmeas includes the barycentric
                velocity of the observatory.
        ephemeris : Name of Ephemeris to be used. List of Ephemeris as queried by jplephem. Default is DE430.
                    For first use Astropy will download the Ephemeris ( for DE430 ~100MB). Options for ephemeris inputs are
                    ['de432s','de430',
                    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de423_for_mercury_and_venus/de423.bsp',
                    'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp']
        leap_dir : Directory where leap seconds file will be saved and maintained (STRING). Eg. '/Users/abc/home/savehere/'. Default is
            script directory.
        leap_update : If True, when the leap second file is more than 6 months old will attempt to download a new one.
                      If False, then will just give a warning message. Default is True.
        SolSystemTarget : When running barycentric correction for a stellar target, Target = None. Default value = None
                To correct for Solar RV observations set Target = 'Sun', for reflected light observations, see below.

                For Solar observations:
                Only inputs required are -
                    1. JDUTC
                    2. loc
                    3. zmeas
                    4. ephemeris
                    5. leap_dir
                    6. leap_update
                    7. predictive

                For Reflected light observations:
                    For observations of the Moon: SolSystemTarget='301' and HorizonsID_type='majorbody'
                    For observations of asteroids: SolSystemTarget=ObjectName, where the ObjectName is queriable by Horizons and HorizonsID_type='smallbody' [default]

        HorizonsID_type : Refers to the Horizons id type to identify the object type, and is required for the reflected light observations.
        > 'smallbody' refers to asteroid or comet and is the default. Use for Asteroids.
        > 'majorbody' refers to planets or satellites. Use for Moon.

        predictive : If True, then instead of returning v_true, returns v_predicted.
                Default: False, and return is v_true from Wright and Eastman (2014)



    See OUTPUTs for description
        FOR STELLAR OBSERVATIONS (not the Sun)
        If SolSystemTarget == None
            If predictive = False
            weighted_vtrue : The barycenter-corrected RV (m/s) for the exposure as weighted by the flux from the exposure meter as defined in Wright & Eastman, 2014.
                    NOTE: This is not just the barycentric velocity that can be subtracted directly from the measured RV.
                    The measured RV must be entered in the code as zmeas. This is because the relativistic cross product between zbary and zmeas is
                    required. This matters at ~ m/s level and hence must be included.

                The formula used is ztrue = ((1.+zb)*(1.+zmeas)-1.)
                Therefore if zmeas is set to 0, then ztrue = zb. The velocities are just the redshift (z) x speed of light (c).

            Else if predictive = True
                weighted_vpredicted : In this predictive mode, the expected RV shift measured by the instrument is output.
                NOTE: zpredicted = zmeas for an ideal noiseless observation, where the RV from the star (ztrue) is zero (or a flat line vs time)

            JDUTCMID : The flux weighted midpoint time is returned.
            warning : Warning and Error message from the routine.
            status : Status regarding warning and error message. Returns the following -
                    0 - No warning or error.
                    1 - Warning message.
                    2 - Error message.

        FOR SOLAR OBSERVATIONS
        If SolSystemTarget == 'Sun'
            If predictive = False
            weighted_v_true: Flux weighted true radial velocity of the Sun for an observer at the observatory but in an inertial frame not moving.
                    If zmeas is included to show the measured absolute redshift for the Sun as measured by an instrument,
                    then in this formulation, v_true will show the motion of the Sun,
                    which is mostly dominated by the synodic period of Jupiter as seen from Earth.
            The formula used is ztrue = ((1.+zb)*(1.+zmeas)-1.)
            Therefore if zmeas is set to 0, then ztrue = zb. The velocities are just the redshift (z) x speed of light (c).

            Else if predictive = True
            weighted_v_predicted: Flux weighted ideal redshift measured for the Sun from Earth for given location and time.
                    This output returns the theoretical prediction for the redshift which includes the barycentric component.
                    This will be the measurement of a noiseless RV instrument observing the Sun.

            JDUTCMID : The flux weighted midpoint time is returned.
            warning : Warning and Error message from the routine.
            status : Status regarding warning and error message. Returns the following -
                    0 - No warning or error.
                    1 - Warning message.
                    2 - Error message.


    '''
    expmeterflux = np.array(expmeterflux)
    error = []

    if isinstance(JDUTC,(int,float)):
        JDUTC = np.array([JDUTC])

    ## Check for size of Flux Array ##
    if len(JDUTC)!=len(expmeterflux):
        print('Error: Size of JDUTC array is not equal to expmeterflux (Flux) array')
        error += ['Error: Size of JDUTC array is not equal to expmeterflux (Flux) array']


    ## Calculate barycentric velocity at each instance of exposure meter reading ##
    vel,warning,status = get_BC_vel(JDUTC=JDUTC,
                                starname=starname,hip_id=hip_id,ra=ra,dec=dec,epoch=epoch,pmra=pmra,pmdec=pmdec,px=px,
                                obsname=obsname,lat=lat,longi=longi,alt=alt,
                                rv=rv,zmeas=zmeas,ephemeris=ephemeris,leap_dir=leap_dir,leap_update=leap_update,
                                SolSystemTarget=SolSystemTarget, HorizonsID_type=HorizonsID_type, predictive=predictive)



    ## Weight it by flux ##

    weighted_vel = flux_weighting(flux = expmeterflux, qty = vel)
    try:
        JDUTC = JDUTC.jd
    except:
        JDUTC = np.array(JDUTC)

    JDUTCMID = flux_weighting(flux=expmeterflux, qty=JDUTC)

    # Error handling
    if error:
        warning.append(error)
        status = 2

    return weighted_vel, JDUTCMID, warning, status



# This is taken from PINT - https://github.com/nanograv/PINT/blob/master/pint/erfautils.py. For licensing see license file on PINT github repository.
# https://github.com/nanograv/PINT/blob/ea1c3b9f71c06a73d6613d647770b99924f7c9bf/LICENSE.md
from __future__ import division
import numpy as np
import astropy.units as u
import datetime
try:
    import astropy.erfa as erfa
except ImportError:
    import astropy._erfa as erfa
import astropy.table as table
from astropy.time import Time
from astropy.utils.iers import conf,IERS_A, IERS_A_URL, IERS_B, IERS_B_URL, IERS
from astropy.utils.data import download_file
from .PhysicalConstants import *


conf.auto_max_age=15

iers_b_file = download_file(IERS_B_URL, cache=True)
iers_b = IERS_B.open(iers_b_file)
IERS.iers_table = iers_b
iers_tab = IERS.iers_table

# Earth rotation rate in radians per UT1 second
#
# This is from Capitaine, Guinot, McCarthy, 2000 and is
# in IAU Resolution B1.8 on the Earth Rotation Angle (ERA)
# and the relation of it to UT1.  The number 1.00273781191135448
# below is a defining constant.  See here:
# http://iau-c31.nict.go.jp/pdf/2009_IAUGA_JD6/JD06_capitaine_wallace.pdf
OM = 1.00273781191135448 * 2.0 * np.pi / SECS_PER_DAY

# arcsec to radians
asec2rad = np.pi/(180*3600)


def gcrs_posvel_from_itrf(loc, toas,tts):
    """Return a list of PosVel instances for the observatory at the TOA times.
    Observatory location should be given in the loc argument as an astropy
    EarthLocation object. This location will be in the ITRF frame (i.e.
    co-rotating with the Earth).

    This routine returns a list of PosVel instances, containing the
    positions (m) and velocities (m / s) at the times of the toas and
    referenced to the Earth-centered Inertial (ECI, aka GCRS) coordinates.
    This routine is basically SOFA's pvtob() [Position and velocity of
    a terrestrial observing station] with an extra rotation from c2ixys()
    [Form the celestial to intermediate-frame-of-date matrix given the CIP
    X,Y and the CIO locator s].


    EDIT:
        Shubham Kanodia 1st December 2017
        Use TT times from Barycorrpy leap second management routine (utc_tdb) and not Astropy.

        tts : Time Object in TT format. Generated by utc_tdb.JDUTC_to_JDTDB()


    """
    # If the input is a single TOA (i.e. a row from the table),
    # then put it into a list
    if type(toas) == table.row.Row:
        ttoas = Time([toas['mjd']])
        N = 1
    elif type(toas) == table.table.Table:
        N = len(toas)
        ttoas = toas['mjd']
    else:
        if toas.isscalar:
            ttoas = Time([toas])
        else:
            ttoas = toas
        N = len(ttoas)



    # Get various times from the TOAs as arrays
    #tts = np.asarray([(t.jd1, t.jd2) for t in ttoas.tt]).T
    tts = np.asarray([tts.jd1,tts.jd2])  # Use Leap second management routine from Barycorrpy

    ut1s = np.asarray([(t.jd1, t.jd2) for t in ttoas.ut1]).T
    mjds = np.asarray(ttoas.mjd)

    # Get x, y coords of Celestial Intermediate Pole and CIO locator s
    X, Y, S = erfa.xys00a(*tts)

    # Get dX and dY from IERS B in arcsec and convert to radians
    dX = np.interp(mjds, iers_tab['MJD'].value, iers_tab['dX_2000A'].value) * asec2rad
    dY = np.interp(mjds, iers_tab['MJD'].value, iers_tab['dY_2000A'].value) * asec2rad

    # Get GCRS to CIRS matrices
    rc2i = erfa.c2ixys(X+dX, Y+dY, S)

    # Gets the TIO locator s'
    sp = erfa.sp00(*tts)

    # Get X and Y from IERS A in arcsec and convert to radians
    #xp = np.interp(mjds, iers_tab['MJD'], iers_tab['PM_X_B']) * asec2rad
    #yp = np.interp(mjds, iers_tab['MJD'], iers_tab['PM_Y_B']) * asec2rad
    # Get X and Y from IERS B in arcsec and convert to radians
    xp = np.interp(mjds, iers_tab['MJD'].value, iers_tab['PM_x'].value) * asec2rad
    yp = np.interp(mjds, iers_tab['MJD'].value, iers_tab['PM_y'].value) * asec2rad

    # Get the polar motion matrices
    rpm = erfa.pom00(xp, yp, sp)

    # Observatory geocentric coords in m
    xyzm = np.array([a.to(u.m).value for a in loc.geocentric])
    x, y, z = np.dot(xyzm, rpm).T

    # Functions of Earth Rotation Angle
    theta = erfa.era00(*ut1s)
    s, c = np.sin(theta), np.cos(theta)
    sx, cx = s * x, c * x
    sy, cy = s * y, c * y

    # Initial positions and velocities
    iposs = np.asarray([cx - sy, sx + cy, z]).T
    ivels = np.asarray([OM * (-sx - cy), OM * (cx - sy), \
                        np.zeros_like(x)]).T
    # There is probably a way to do this with np.einsum or something...
    # and here it is .
    poss = np.empty((N, 3), dtype=np.float64)
    vels = np.empty((N, 3), dtype=np.float64)
    poss = np.einsum('ij,ijk->ik', iposs, rc2i)
    vels = np.einsum('ij,ijk->ik', ivels, rc2i)
    return poss,vels



from __future__ import division
from __future__ import print_function
#import de423    #https://pypi.python.org/pypi/jplephem
from astropy.coordinates import EarthLocation
from astropy.coordinates import get_body_barycentric_posvel, get_body_barycentric
from astropy.time import Time
from astroquery.jplhorizons import Horizons

import astropy.constants as ac
import numpy as np
import os


from . import find_hip
from . import PINT_erfautils as PINT
from . import utc_tdb
from .utils import flux_weighting, get_stellar_data, CalculatePositionVector
from .PhysicalConstants import *


def SolarBarycentricCorrection(JDUTC, loc, zmeas=0, ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True, predictive=False):
    """
    Perform barycentric correction for Solar observations.
    In the solar case, ignores retarded time, i.e. the change in position angle of the Sun wrt the observatory in the 8 minutes it takes for the light to reach Earth.
    This is because the Solar velocity is low enough (~ 10m/s), that in 8 mins the change in position over 1 AU is negligible.
    INPUTS:
        JDUTC: Astropy Time object in UTC scale with JD format
        loc: Astropy EarthLocation Object)
        zmeas : Measured redshift (e.g., the result of cross correlation with template spectrum). Default is 0.
                The redshift measured by the spectrograph before any barycentric correction. Therefore zmeas includes the barycentric
                velocity of the observatory.
        ephemeris: Name of Ephemeris to be used. List of Ephemeris as queried by jplephem. Default is DE430.
                For first use Astropy will download the Ephemeris ( for DE430 ~100MB). Options for ephemeris inputs are
                ['de432s','de430',
                'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de423_for_mercury_and_venus/de423.bsp',
                'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp']
        leap_dir: Directory where leap seconds file will be saved and maintained (STRING). Eg. '/Users/abc/home/savehere/'. Default is
                script directory.
        leap_update: If True, when the leap second file is more than 6 months old will attempt to download a new one.
                If False, then will just give a warning message. Default is True.

        predictive : If True, then instead of returning v_true, returns v_predicted.
        Default: False, and return is v_true from Wright and Eastman (2014)

        See OUTPUTs for description

    OUTPUTS:
            If predictive = False
                v_true: The true radial velocity of the Sun for an observer at the observatory but in an inertial frame not moving.
                    If zmeas is included to show the measured absolute redshift for the Sun as measured by an instrument,
                    then in this formulation, v_true will show the motion of the Sun,
                    which is mostly dominated by the synodic period of Jupiter as seen from Earth.

            Else if predictive = True
                v_predicted: Ideal redshift measured for the Sun from Earth for given location and time.
                    This output returns the theoretical prediction for the measured redshift which includes the barycentric component.
                    This will be the measurement of a noiseless RV instrument observing the Sun.


        The formula used is ztrue = ((1.+ zb)*(1.+ zmeas)-1.)
        Therefore if zmeas is set to 0, then ztrue = zb. The velocities are just the redshift (z) x speed of light (c).


    """

    # Convert times to obtain TDB and TT

    JDTDB, JDTT, warning, error = utc_tdb.JDUTC_to_JDTDB(JDUTC, fpath=leap_dir, leap_update=leap_update)

    ##### NUTATION, PRECESSION, ETC. #####

    r_pint, v_pint = PINT.gcrs_posvel_from_itrf(loc, JDUTC, JDTT)

    r_eci = r_pint[0]  # [m]
    v_eci = v_pint[0]  # [m/s]

    ##### EPHEMERIDES #####

    earth_geo = get_body_barycentric_posvel('earth', JDTDB, ephemeris=ephemeris) # [km]
    PosVector_EarthSSB = r_eci + earth_geo[0].xyz.value*1000. # [m]
    v_geo = earth_geo[1].xyz.value*1000./86400.  # [m/s]

    # Relativistic Addition of Velocities
    VelVector_EarthSSB = (v_eci+v_geo) / (1.+ np.sum(v_eci*v_geo)/c**2) # [m/s]
    BetaEarth = VelVector_EarthSSB / c

    GammaEarth = 1. / np.sqrt(1.- np.sum(BetaEarth**2))

    ## Ignoring retarded time for Solar vectors

    solar_ephem = get_body_barycentric_posvel('sun', JDTDB, ephemeris=ephemeris)

    PosVector_SolSSB = solar_ephem[0].xyz.value*1000. #[m]
    VelVector_SolSSB = solar_ephem[1].xyz.value*1000./86400.  # [m/s]
    BetaSolar = VelVector_SolSSB / c

    GammaSolar = 1. / np.sqrt(1.- np.sum(BetaSolar**2))



    PosVector_SolEarth, PosMag_SolEarth, PosHat_SolEarth = CalculatePositionVector(r1=PosVector_SolSSB, r2=PosVector_EarthSSB)

    # To correct for redshift experienced by photon due to Earth's gravity well,
    # also as it travels further in the Sol System from the Sun to the position of the Earth (Solar gravity well)
    # Correcting for this should give a blue shift.
    zGREarth =   - ac.G.value * ac.M_earth.value / ((ac.c.value**2)*(np.sqrt(np.sum(r_eci**2))))\
                - ac.G.value * ac.M_sun.value / ((ac.c.value**2)*(np.sqrt(np.sum(PosVector_SolEarth**2))))

    # To correct for redshift experienced by photon and place it in the SSB
    zGRSun =  - (ac.G.value * ac.M_sun.value) / ((ac.c.value**2) * (np.sqrt(np.sum(ac.R_sun.value**2))))

    zpredicted= ( (GammaSolar * (1 + np.dot(BetaSolar,PosHat_SolEarth))*(1+zGREarth)) /
                (GammaEarth * (1 + np.dot(BetaEarth,PosHat_SolEarth)) * (1+zGRSun))  ) - 1

    zclassical = (1 + np.dot(BetaSolar,PosHat_SolEarth))/(1 + np.dot(BetaEarth,PosHat_SolEarth))-1

    zb = ((GammaEarth*(1 + np.dot(BetaEarth,PosHat_SolEarth))) / (1+zGREarth)) - 1

    v_true = c * ((1.+zb)*(1.+ zmeas)-1.)  # [m/s]
    v_predicted = c * zpredicted # [m/s]
#
#     print('JDTDB=',JDTDB.jd)
#     print('PosVector_SolSSB=', PosVector_SolSSB)
#     print('VelVector_SolSSB=', VelVector_SolSSB)
#     print('PosVector_EarthSSB=', PosVector_EarthSSB)
#     print('VelVector_EarthSSB=', VelVector_EarthSSB)
#     print('rhohat=', PosHat_SolEarth)
#     print('GammaSolar=',GammaSolar)
#     print('GammaEarth=',GammaEarth)
#     print('zGRSun=', zGRSun)
#     print('zGREarth=', zGREarth)
#     print('zGR=', ((1+zGREarth)/(1+zGRSun) - 1),'=', ((1+zGREarth)/(1+zGRSun) - 1)*c)
#     print('zSR=', 1 - GammaEarth/GammaSolar)
#     print('Classical=',zclassical, '=', zclassical*c)


    if predictive:
        vel = v_predicted
    else:
        vel = v_true

    return vel, warning, error


def ReflectedLightBarycentricCorrection(SolSystemTarget, JDUTC, loc, zmeas=0, HorizonsID_type='smallbody',
    ephemeris='de430', leap_dir=os.path.join(os.path.dirname(__file__),'data'), leap_update=True, predictive=False):
    """
    Computing the barycentric corrections for reflected light observations of a target in the Solar system.

    INPUTS:
        TargetName
        JDUTC: Astropy Time object in UTC scale with JD format
        loc: Astropy EarthLocation Object)
        zmeas : Measured redshift (e.g., the result of cross correlation with template spectrum). Default is 0.
                The redshift measured by the spectrograph before any barycentric correction. Therefore zmeas includes the barycentric
                velocity of the observatory.
        ephemeris: Ephemeris used only for calculating the position and velocity vectors of the Earth wrt the Solar System Barycenter (SSB).
                For reflected light target and the Sun, the HORIZONS query is used from Astroquery.
                Name of Ephemeris to be used. List of Ephemeris as queried by jplephem. Default is DE430.
                For first use Astropy will download the Ephemeris ( for DE430 ~100MB). Options for ephemeris inputs are
                ['de432s','de430',
                'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de423_for_mercury_and_venus/de423.bsp',
                'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp']
        leap_dir: Directory where leap seconds file will be saved and maintained (STRING). Eg. '/Users/abc/home/savehere/'. Default is
                script directory.
        leap_update: If True, when the leap second file is more than 6 months old will attempt to download a new one.
                If False, then will just give a warning message. Default is True.

        predictive : If True, then instead of returning v_true, returns v_predicted.
            Default: False, and return is v_true from Wright and Eastman (2014)

        SolSystemTarget : When running barycentric correction for a stellar target, Target = None. Default value = None
                To correct for Solar RV observations set Target = 'Sun', for reflected light observations, see below.

                For Reflected light observations:
                    For observations of the Moon: SolSystemTarget='301' and HorizonsID_type='majorbody'
                    For observations of asteroids: SolSystemTarget=ObjectName, where the ObjectName is queriable by Horizons and HorizonsID_type='smallbody' [default]

        HorizonsID_type : Refers to the Horizons id type to identify the object type, and is required for the reflected light observations.
        > 'smallbody' refers to asteroid or comet and is the default. Use for Asteroids.
        > 'majorbody' refers to planets or satellites. Use for Moon.

        See OUTPUTs for description

    OUTPUTS:
            If predictive = False
                v_true: The true radial velocity of the SolSystemTarget for an observer at the observatory but in an inertial frame not moving.
                    If zmeas is included to show the measured absolute redshift for the SolSystemTarget as measured by an instrument,
                    then in this formulation, v_true will show the motion of the SolSystemTarget.
                    Else if predictive = True
                v_predicted: Ideal redshift measured for the SolSystemTarget from Earth for given location and time.
                    This output returns the theoretical prediction for the redshift which includes the barycentric component.
                    This will be the measurement of a noiseless RV instrument observing the SolSystemTarget.


    """


    # Convert times to obtain TDB and TT
    JDTDB, JDTT, warning, error = utc_tdb.JDUTC_to_JDTDB(JDUTC, fpath=leap_dir, leap_update=leap_update)


    # Need dictionary object for HORIZONS call
    # longi = loc.lon.degree
    # lat = loc.lat.degree
    # alt = loc.height.value
    # loc_dict = {'lon': longi,
    #             'lat': lat,
    #             'elevation': alt}

    # Reflecting object
    # First find the light travel time for Object with respect to observatory.

    # try:
    #     TargetObj1 = Horizons(id=SolSystemTarget, location=loc_dict, epochs=JDTDB, id_type=HorizonsID_type).vectors()
    # except ValueError:
    #     warning+= ['Unable to use Vector query for Horizons search using exact observatory coordinates, reverting to using Geocenter. ']
    try:
        TargetObj1 = Horizons(id=SolSystemTarget, location='399', epochs=JDTDB, id_type=HorizonsID_type).vectors()
    except Exception as e:
        print(e)
        raise Exception("Error with Horizons call, id={}, location='399' (Geocenter), epochs={}, id_type={}".format(SolSystemTarget, JDTDB, HorizonsID_type))

    # Here we ignore the light travel time from Geocenter to Earth's surface (observatory)
    EarthTargetLightTravel = TargetObj1['lighttime'][0] #days

    # Subtract light time and find pos and vel for target wrt SSB
    TargetObjTime = JDTDB - (EarthTargetLightTravel)
    TargetObj2 = Horizons(id=SolSystemTarget, location='@0', epochs=TargetObjTime, id_type=HorizonsID_type)
    TargetVectors = TargetObj2.vectors(refplane='earth')
    TargetSSBLightTravel = TargetVectors['lighttime'][0] #days

    [TargetVectors[i].convert_unit_to('m') for i in ['x','y','z']] # Convert from AU to km
    [TargetVectors[i].convert_unit_to('m/s') for i in ['vx','vy','vz']] # Convert from AU/d to km/d
    PosVector_TargetSSB = TargetVectors['x','y','z'].as_array()[0].view((float,3))
    VelVector_TargetSSB = TargetVectors['vx','vy','vz'].as_array()[0].view((float,3))
    BetaTarget = VelVector_TargetSSB / c

    ## SUN
    # First find the light travel time for Sun with respect to Observatory
    #SolObj1 = Horizons(id='Sun', location='@0', epochs=JDTDB, id_type='majorbody').ephemerides()
    #SolSSBLightTravel = SolObj1['lighttime'][0] # Units of days


    # Ignoring difference in light travel time between the Sun and SSB for finding the Solar vectors
    # Subtract light time and find pos and vel for Sun wrt SSB
    TargetSolLightTravelDelay = EarthTargetLightTravel + TargetSSBLightTravel # The
    SolObj2 = Horizons(id='Sun', location='@0', epochs=JDTDB-(TargetSolLightTravelDelay), id_type='majorbody')
    SolVectors = SolObj2.vectors(refplane='earth')

    [SolVectors[i].convert_unit_to('m') for i in ['x','y','z']]
    [SolVectors[i].convert_unit_to('m/s') for i in ['vx','vy','vz']]
    PosVector_SolSSB = SolVectors['x','y','z'].as_array()[0].view((float,3))
    VelVector_SolSSB = SolVectors['vx','vy','vz'].as_array()[0].view((float,3))
    BetaSolar = VelVector_SolSSB / c
    GammaSolar = 1. / np.sqrt(1.- np.sum(BetaSolar**2))


    ##### NUTATION, PRECESSION, ETC. #####
    r_pint, v_pint = PINT.gcrs_posvel_from_itrf(loc, JDUTC, JDTT)

    r_eci = r_pint[0]  # [m]
    v_eci = v_pint[0]  # [m/s]

    earth_geo = get_body_barycentric_posvel('earth', JDTDB, ephemeris=ephemeris) # [km]
    PosVector_EarthSSB = r_eci + earth_geo[0].xyz.value*1000. # [m]
    v_geo = earth_geo[1].xyz.value*1000./86400.  # [m/s]
    VelVector_EarthSSB = (v_eci+v_geo) / (1.+ np.sum(v_eci*v_geo)/c**2) # [m/s]

    BetaEarth = VelVector_EarthSSB / c
    GammaEarth = 1. / np.sqrt(1. - np.sum(BetaEarth**2))


    PosVector_SolEarth, PosMag_SolEarth, PosHat_SolEarth = CalculatePositionVector(r1=PosVector_SolSSB, r2=PosVector_EarthSSB)
    PosVector_SolTarget, PosMag_SolTarget, PosHat_SolTarget = CalculatePositionVector(PosVector_SolSSB, PosVector_TargetSSB)
    PosVector_TargetEarth, PosMag_TargetEarth, PosHat_TargetEarth = CalculatePositionVector(PosVector_TargetSSB, PosVector_EarthSSB)

    # Calculates the the redshift experienced by photon due to Earth's + Sun's gravity well from infinity.
    # Correcting for this should give a blue shift.
    zGREarth =  - ac.G.value * ac.M_earth.value / ((ac.c.value**2)*(np.sqrt(np.sum(r_eci**2))))\
                - ac.G.value * ac.M_sun.value / ((ac.c.value**2)*(np.sqrt(np.sum(PosVector_SolEarth**2))))

    # To correct for redshift experienced by photon and place it in the SSB
    zGRSun =  - (ac.G.value * ac.M_sun.value) / ((ac.c.value**2) * (np.sqrt(np.sum(ac.R_sun.value**2))))

    zGR = (1+zGREarth)/(1+zGRSun)
    zSR = GammaSolar/GammaEarth
    zClassical1 = (1 + np.dot(BetaSolar, PosHat_SolTarget))/(1 + np.dot(BetaTarget, PosHat_SolTarget))
    zClassical2 =  (1 + np.dot(BetaTarget, PosHat_TargetEarth))/(1 + np.dot(BetaEarth, PosHat_TargetEarth))


    zpredicted = zGR*zSR*zClassical1*zClassical2 - 1

    zb = (GammaEarth * (1+np.dot(BetaEarth, PosHat_TargetEarth)) * \
        (1 + np.dot(BetaTarget, PosHat_SolTarget))/(1 + np.dot(BetaTarget, PosHat_TargetEarth)) \
        / (1+zGREarth)) - 1

    v_true = c * ((1.+zb)*(1.+ zmeas)-1.)  # [m/s]
    v_predicted = c * zpredicted # [m/s]

    if predictive:
        vel = v_predicted
    else:
        vel = v_true

    return vel, warning, error



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()



setup(name='barycorrpy',
      version='0.3.4',
      description='Barycentric Velocity correction at 1 cm/s level',
      long_description=readme(),
      url='https://github.com/shbhuk/barycorrpy',
      author='Shubham Kanodia',
      author_email='shbhuk@gmail.com',
      install_requires=['astropy>=3.2.3','jplephem','numpy>=1.17.2','scipy','astroquery>=0.3.10'],
      packages=['barycorrpy', 'barycorrpy.tests'],
      license='GPLv3',
      classifiers=['Topic :: Scientific/Engineering :: Astronomy'],
      keywords='Barycentric Correction Astronomy Spectroscopy Radial Velocity',
      include_package_data=True
      )



import astropy.constants as ac
import numpy as np


# Parsing constants #
AU = ac.au.value # [m]
c = ac.c.value # Speed of light [m/s]
pctoau = 3600*180/np.pi # No. of AU in one parsec
SECS_PER_DAY = 3600 * 24
year = 365.25*SECS_PER_DAY # [s]
kmstoauyr = 1000 * year/(AU) #km/s to AU/year

M_moon = 73476730924573500000000 # Mass of the Moon [kg]

# Mass and normalised mass of solar system bodies
ss_bodies = ['Sun', 'Mercury', 'Venus', 'Earth', 'Mars', 'Jupiter', 'Saturn', 'Uranus', 'Neptune', 'Moon']
M = dict(zip(ss_bodies, [ac.M_sun.value, 0.3301e24, 4.867e24, ac.M_earth.value, 0.6417e24, ac.M_jup.value, 568.5e24, 86.82e24, 102.4e24, M_moon])) # [kg]

#Reduced number of bodies, to increase speed. Does not affect precision at 1 cm/s level
ss_bodies = ['Sun','Earth','Jupiter','Saturn']
M = dict(zip(ss_bodies, [ac.M_sun.value, ac.M_earth.value,ac.M_jup.value,568.5e24])) # [kg]

GM = {k:ac.G.value*M[k] for k in ss_bodies}



from __future__ import print_function
import numpy as np
try:
    from astroquery.simbad import Simbad
except:
    print('Cannot import astroquery.simbad')
import astropy.units as u
from astropy.coordinates import SkyCoord
import os
# import winsound
# frequency = 2000  # Set Frequency To 2500 Hertz
# duration = 500 

def flux_weighting(flux,qty):
    '''
    INPUT:
        flux - Numpy Array. Will weight qty array by flux
        qty - Quantity to be normalized / weighted.
    
    OUTPUT:
        Normalized value
    '''
    
    return sum(qty * flux)/sum(flux)
    
def get_stellar_data(name=''):
    '''
    Function to query Simbad for following stellar information RA, Dec, PMRA, PMDec, Parallax Epoch
    INPUTS:
        name = Name of source. Example 
    
    
    '''
    warning = []
    
    customSimbad = Simbad()
    customSimbad.add_votable_fields('ra(2;A;ICRS;J2000)', 'dec(2;D;ICRS;J2000)','pm', 'plx','parallax','rv_value')
    #Simbad.list_votable_fields()
    customSimbad.remove_votable_fields( 'coordinates')
    #Simbad.get_field_description('orv')
    obj = customSimbad.query_object(name)
    if obj is None:
        raise ValueError('ERROR: {} target not found. Check target name or enter RA,Dec,PMRA,PMDec,Plx,RV,Epoch manually\n\n'.format(name))
    else:        
        warning += ['{} queried from SIMBAD.'.format(name)]

    # Check for masked values
    if all([not x for x in [obj.mask[0][i] for i in obj.colnames]])==False:
        warning += ['Masked values present in queried dataset']


    obj = obj.filled(None)
    
    pos = SkyCoord(ra=obj['RA_2_A_ICRS_J2000'],dec=obj['DEC_2_D_ICRS_J2000'],unit=(u.hourangle, u.deg))
    ra = pos.ra.value[0]
    dec = pos.dec.value[0]
    pmra = obj['PMRA'][0]
    pmdec = obj['PMDEC'][0]
    plx = obj['PLX_VALUE'][0]
    rv = obj['RV_VALUE'][0] * 1000 #SIMBAD output is in km/s. Converting to m/s
    epoch = 2451545.0
    
    star = {'ra':ra,'dec':dec,'pmra':pmra,'pmdec':pmdec,'px':plx,'rv':rv,'epoch':epoch}
    
    # Fill Masked values with None. Again. 
    for i in star:
        if star[i] > 1e10:
            star[i] = None           
       
    warning += ['Values queried from SIMBAD are {}'.format(star)]
  
    
    return star,warning
    
def find_hip(hip_index,cat_dir=os.path.join(*[os.path.dirname(__file__),'data','hip2.dat'])):
    '''
    NOTE: Hipparcos Catalogue Epoch is J1991.25 or JD 2448349.0625
    
    INPUT:
        hip_index : The index of the star that needs to be searched.
        cat_dir : Directory where catalogue is saved. 
    OUTPUT:
        hip_id : Index of Star
        ra : RA of star in degrees
        dec : Declination of star in degrees
        px_mas : Parallax angle in milli-arcseconds
        pmra : Proper motion in RA in milli-arcseconds/year
        pmdec : Proper motion in Declination in milli-arcseconds/year 
        epoch : Epoch of Catalogue - J1991.25 , JD 2448349.0625       
    
    '''
    epoch = 2448349.0625
    
    hip_id=[]
    ra=[]
    dec=[]
    px_mas=[]
    pmra=[]
    pmdec=[]

    
    with open(cat_dir) as f:
        for line in f:
            a = line.split()
            hip_id.append(float(a[0]))
            ra.append((float(a[4])*180.)/np.pi) # Convert to degrees
            dec.append((float(a[5])*180.)/np.pi) # Convert from radians to degrees
            px_mas.append(float(a[6]))  # in mas
            pmra.append(float(a[7])) # in mas/year
            pmdec.append(float(a[8]))  # in mas/year
        
    index=np.where(np.array(hip_id)==hip_index)[0][0]
    
    star = {'ra':ra[index],'dec':dec[index],'pmra':pmra[index],'pmdec':pmdec[index],'px':px_mas[index],'epoch':epoch}
    
    return star

      
def CalculatePositionVector(r1, r2):
    '''
    INPUTS:
        r1, r2 = (x,y,z) positions for Obj 1 and Obj 2.
    OUTPUTS:    
        vec(X) = vec(r1) - vec(r2)
        Xmag, Xhat = Magnitude of the position vector, and its unit vector resp. are also returned.
    
    '''
    # Vector from object barycenter to Observatory
    X = np.array(r1-r2)
    Xmag = np.sqrt(np.sum(X*X)) # [m]
    Xhat = X / Xmag # Unitless
    
    return X, Xmag, Xhat


from __future__ import print_function
from __future__ import division
from astropy.time import Time
import numpy as np
from .barycorrpy import get_BC_vel , exposure_meter_BC_vel
from . import utc_tdb

def run_sample():
            a=[]
            b=0

            JDUTC = 2458000 # Also accepts float input for JDUTC. Verify scale and format

            # Observation of Tau Ceti taken from CTIO on JD 2458000.
            # Observatory location manually entered. Stellar positional parameters taken from Hipparcos Catalogue
            result = get_BC_vel(JDUTC=JDUTC, hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9, ephemeris='de430', zmeas=0.0)

            if np.isclose(a=result[0], b=15403.9508, atol=1e-2, rtol=0):
                a.append('result')
                b+=1


            # Observation of Tau Ceti taken from CTIO on JD 2458000. Observatory location taken from Astropy list.
            # Stellar positional parameters taken from Hipparcos Catalogue
            JDUTC = Time(2458000, format='jd', scale='utc')
            result2  = get_BC_vel(JDUTC=JDUTC, hip_id=8102, obsname='CTIO', ephemeris='de430')

            if np.isclose(a=result2[0], b=15403.9608, atol=1e-2, rtol=0):
                a.append('result2')
                b+=1

            # Observation of Tau Ceti taken from CTIO on JD 2458000,2458010,2458020.
            # Observatory and stellar parameters entered by user.
            # Use DE405 ephemeris

            ra=26.0213645867
            dec=-15.9395557246
            obsname=''
            lat=-30.169283
            longi=-70.806789
            alt=2241.9
            epoch = 2451545.0
            pmra = -1721.05
            pmdec = 854.16
            px = 273.96
            rv = 0.0
            zmeas=0.0
            JDUTC=[2458000,2458000.00001,2458000.00002] # Can also enter JDUTC as float instead of Astropy Time Object
            ephemeris='https://naif.jpl.nasa.gov/pub/naif/generic_kernels/spk/planets/a_old_versions/de405.bsp'

            result3=get_BC_vel(JDUTC=JDUTC, ra=ra, dec=dec, obsname=obsname, lat=lat, longi=longi, alt=alt, pmra=pmra,
                pmdec=pmdec, px=px, rv=rv, zmeas=zmeas,epoch=epoch, ephemeris=ephemeris, leap_update=True)

            if np.allclose([result3[0][0],result3[0][1],result3[0][2]],[15407.4860,15407.4723,15407.4586],atol=1e-2, rtol=0):
                a.append('result3')
                b+=1



            # Exposure meter calculation
            flux = [4.5,1.5,2] # Same number of elements as JDUTC

            result4,JDUTCMID,warning4,status4=exposure_meter_BC_vel(JDUTC=JDUTC, expmeterflux = flux,
                ra=ra, dec=dec, obsname=obsname, lat=lat, longi=longi, alt=alt, pmra=pmra,
                pmdec=pmdec, px=px, rv=rv, zmeas=zmeas, epoch=epoch, ephemeris=ephemeris, leap_update=True)

            if np.isclose(a=result4, b=15407.4765, atol=1e-2, rtol=0):
                a.append('result4')
                b+=1

            # JDUTC to BJDTDB time converter
            corr_time = utc_tdb.JDUTC_to_BJDTDB(JDUTC, hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9)

            if np.allclose([corr_time[0]], [2458000.00505211,  2458000.00506211,  2458000.00507211], atol = 1e-7, rtol = 0):
                a.append('corr_time')
                b+=1

            # Predictive Mode
            result5 = get_BC_vel(JDUTC=2458000, hip_id=8102, lat=-30.169283, longi=-70.806789, alt=2241.9, ephemeris='de430', zmeas=0.0, predictive=True)

            if np.isclose(a=result5[0], b=-15403.15938, atol=1e-2, rtol=0):
                a.append('result5')
                b+=1

            result6 = get_BC_vel(JDUTC=2458000, lat=-30.169138888, longi=-70.805888, alt=2379.5, zmeas=0.0, SolSystemTarget='Sun')

            if np.isclose(a=result6[0], b=819.4474, atol=1e-2, rtol=0):
                a.append('result6')
                b+=1


            if b==7:
                print('***********SUCCESS**************\nAll barycentric correction velocities match expected values to 1 cm/s\n')
            else:
                print('{} out of 7 results match. Compare outputs vs those on the github wiki. Check others - \n'.format(b,a))

            return result, result2, result3, result4, JDUTCMID, warning4, status4, corr_time, result5, result6
